#include <hardware/pwm.h>
#include <LittleFS.h>

//44k 22
#define TIMER1_INTERVAL_us 22
//32k 31
//#define TIMER1_INTERVAL_us 31
//16k 62
//#define TIMER1_INTERVAL_us 62
//8k 125
//#define TIMER1_INTERVAL_us 125

#define PIN_BUZZER    D4
#define PIN_DEBUG_PWM    D12
#define PIN_DEBUG_MANAGE    D11

//bufferの大きさ
//#define BUF_LENGTH 32768
//#define BUF_LENGTH 16384
//#define BUF_LENGTH 8192
#define BUF_LENGTH 4096
//#define BUF_LENGTH 2048
//#define BUF_LENGTH 1024

//和音数
#define WAVE_CHORDS 8

//waveのオフセット Waveファイルの始めに43個のヘッダー領域があるため、そこを飛ばす。
#define WAVE_OFFSET 44

//wave_modeのenum
#define WAVE_MODE_STOP 0
#define WAVE_MODE_RUN_START 1
#define WAVE_MODE_RUNNING 2
#define WAVE_MODE_RUN_END 3
#define WAVE_MODE_RUN_REPEAT 4
#define WAVE_MODE_NONE 5

//8bitの時127，16bitの時は0 無音状態
#define WAVE_DEFAULT_VALUE 127

#define BUF_CH_A 0
#define BUF_CH_B 1

//timer　オブジェクト
repeating_timer_t timer;

//各chの音　そのうち構造体にするか？
uint32_t wave_position[WAVE_CHORDS];//現在値
uint32_t wave_length[WAVE_CHORDS];//長さ
uint16_t wave_volume[WAVE_CHORDS];//volume 0-255
String wave_name[WAVE_CHORDS];//ファイル名
String wave_next_name[WAVE_CHORDS];//ファイル名
uint8_t wave_last_val[WAVE_CHORDS];//プチ音を消すため最後の音を保つための値。
int wave_status[WAVE_CHORDS];//状態　WAVE_MODE_STOP,WAVE_MODE_RUN
int wave_order[WAVE_CHORDS];//要求　WAVE_MODE_STOP,WAVE_MODE_RUN


//出力部分　割込みで使用するので、volatile
volatile uint slice_num = 0;
volatile uint16_t wave_buf_a[BUF_LENGTH];
volatile uint16_t wave_buf_b[BUF_LENGTH];
volatile int buf_ab = 1;
volatile uint16_t buf_count = 0;
volatile bool buf_flag;//false:次のバッファを書かないといけない、true:書き込み済み

//LittelFS用
File f;

//UART用
String inputString = "";         // a String to hold incoming data
bool stringComplete = false;  // whether the string is complete

bool timer_callback( repeating_timer_t *rt ){
  //割込みで呼ばれるところ
  //デバッグ用
  gpio_put( PIN_DEBUG_PWM, HIGH );
  //PWMへの書き込み A,Bで分かれる。
  if(buf_ab == BUF_CH_A){
    pwm_set_chan_level(slice_num, PWM_CHAN_A, wave_buf_a[buf_count]);
  }
  else{
    pwm_set_chan_level(slice_num, PWM_CHAN_A, wave_buf_b[buf_count]);
  }
  buf_count ++;
  //bufferの切り替え
  if(buf_count == BUF_LENGTH){
    if(buf_ab == BUF_CH_A){
      buf_ab = BUF_CH_B;
    }
    else{
      buf_ab = BUF_CH_A;
    }
    buf_flag  = false;
    buf_count  = 0;
  }
  //デバッグ用
  gpio_put( PIN_DEBUG_PWM, LOW );
  return true;
}

void manage_wave_sub(int ch,uint8_t *temp_buf)
{
  //各ch内のStateMachineをバッファ取得まで進める。
  bool buf_flag = false;
  //デフォルト値初期化
  for(int j = 0; j < BUF_LENGTH;j++){
    temp_buf[j] = wave_last_val[ch];
  }
  //バッファが入るまでStateMachineを回す。
  while(!buf_flag){
    //強制終了(どのStateに行っても一緒なのでここに書く）
    if(wave_order[ch] == WAVE_MODE_STOP){
      wave_status[ch] = WAVE_MODE_STOP;
    }
    //StateMachine
    switch(wave_status[ch])
      {
      case WAVE_MODE_STOP:
        //Serial.println("WAVE_MODE_STOP");
        //停止時
        switch(wave_order[ch])
        {
          case WAVE_MODE_STOP:
            //中止状態継続
            wave_status[ch] = WAVE_MODE_STOP;
            buf_flag = true;
            break;
          case WAVE_MODE_RUN_START:
            //RUNへ移行
            wave_status[ch] = WAVE_MODE_RUN_START;
            break;
          default:
            //WAVE_MODE_NONE
            buf_flag = true;
            break;            
        }
        break;
      case WAVE_MODE_RUN_START:
        //Serial.println("WAVE_MODE_RUN_START");
        //ファイルオープンと初期値把握
        f = LittleFS.open(wave_name[ch],"r");      
        wave_length[ch] = f.size() - WAVE_OFFSET;
        wave_position[ch] = 0;
        f.close();
        //RUNNINGへ移行
        wave_status[ch] = WAVE_MODE_RUNNING;        
        break;
      case WAVE_MODE_RUNNING:
        //Serial.println("WAVE_MODE_RUNNING");
        //データ数により振り分け
        //行き先はRUNNNINGとENDとREPEATの三つ
        if(wave_length[ch] - wave_position[ch] < BUF_LENGTH){
          //データがバッファ途中で終了の時
          if(wave_next_name[ch] == "none"){
            //終了のとき
            wave_status[ch] = WAVE_MODE_RUN_END;
          }
          else{
            //Repeatのとき
            wave_status[ch] = WAVE_MODE_RUN_REPEAT;
          }
        }
        else{
          //Buffer分　格納する
          f = LittleFS.open(wave_name[ch],"r");      
          f.seek(wave_position[ch] + WAVE_OFFSET,SeekSet);
          f.read(temp_buf,BUF_LENGTH);
          wave_last_val[ch] = temp_buf[BUF_LENGTH-1];
          wave_position[ch] = f.position() - WAVE_OFFSET;
          f.close();
          wave_status[ch] = WAVE_MODE_RUNNING;
          buf_flag = true;
        }
        break;
      case WAVE_MODE_RUN_END:
        //Serial.println("WAVE_MODE_RUN_END");
        //残りのdataを格納しあとは最終データで埋める。
        f = LittleFS.open(wave_name[ch],"r");      
        f.seek(wave_position[ch] + WAVE_OFFSET,SeekSet);
        f.read(temp_buf,(wave_length[ch] - wave_position[ch]));
        f.close();
        wave_last_val[ch] = temp_buf[wave_length[ch] - wave_position[ch] - 2];
        //残りを同じ値で埋めておく。
        for(int j = wave_length[ch] - wave_position[ch] - 1;j < BUF_LENGTH;j++){
          temp_buf[j] = wave_last_val[ch];
        }
        buf_flag = true;
        wave_status[ch] = WAVE_MODE_STOP;
        break;
      case WAVE_MODE_RUN_REPEAT:
        //Serial.println("WAVE_MODE_RUN_REPEAT");
        //残りのdataを格納する。
        f = LittleFS.open(wave_name[ch],"r");      
        f.seek(wave_position[ch] + WAVE_OFFSET,SeekSet);
        f.read(temp_buf,(wave_length[ch] - wave_position[ch]));
        f.close();
        //bufferの最終書き込み位置を記憶する。
        uint32_t wave_pos;
        wave_pos = wave_length[ch] - wave_position[ch] - 1;
        //次のデータで埋める
        f = LittleFS.open(wave_name[ch],"r");      
        wave_length[ch] = f.size() - WAVE_OFFSET;
        wave_position[ch] = 0;
        f.seek(wave_position[ch] + WAVE_OFFSET,SeekSet);
        //f.read(temp_buf + wave_pos + 1,(BUF_LENGTH - wave_pos));
        f.read(temp_buf + wave_pos,(BUF_LENGTH - wave_pos));
        wave_last_val[ch] = temp_buf[BUF_LENGTH-1];
        wave_position[ch] = f.position() - WAVE_OFFSET;
        f.close();
        /*
        Serial.println("second_data");
        Serial.println(wave_pos);
        Serial.println(BUF_LENGTH - wave_pos);
        for(int i=0;i<BUF_LENGTH;i++){
          Serial.print(i);
          Serial.print(",");
          Serial.println(temp_buf[i]);
        }
        while(true);
        */
        wave_status[ch] = WAVE_MODE_RUNNING;
        buf_flag = true;
        break;
      default:
        //Serial.println("WAVE_MODE_NONE");
        //WAVE_MODE_NONE
        //ここには来ないはずだが。
        break;          
    }
    wave_order[ch] = WAVE_MODE_NONE;
  }
}

void manage_wave()
{
  //バッファへの書き込み
  //書く必要があるかの確認
  if(buf_flag == true){
    return;
  }
  //デバッグ用
  gpio_put( PIN_DEBUG_MANAGE, HIGH );

  //音バッファを０リセット
  if(buf_ab == BUF_CH_A){
    for(int i = 0; i < BUF_LENGTH;i++){
      wave_buf_b[i] = 511;
    }
  }
  else{
    for(int i = 0; i < BUF_LENGTH;i++){
      wave_buf_a[i] = 511;
    }
  }
  
  //buffer　　globalやstaticにした方が初期化を毎回しないので高速。（あとでやる）
  uint8_t temp_buf[BUF_LENGTH];

  //和音回数分繰り返す
  for(int i = 0;i < WAVE_CHORDS;i++){
    //wave_statusで場合分けして、各ch　buffer入力まで処理
    manage_wave_sub(i,temp_buf);
    //BufferのMixとA,Bchマスタバッファへの格納
    //音量変更はそのうち
    if(buf_ab == BUF_CH_A){
      for(int i = 0; i < BUF_LENGTH;i++){
        wave_buf_b[i] += (uint16_t)temp_buf[i];
        wave_buf_b[i] -= 127;
      }
    }
    else{
      for(int i = 0; i < BUF_LENGTH;i++){
        wave_buf_a[i] += (uint16_t)temp_buf[i];
        wave_buf_a[i] -= 127;
      }
    }
    
  }
  buf_flag = true;

  //デバッグ用
  gpio_put( PIN_DEBUG_MANAGE, LOW );

  return;
}

void play_wav(int8_t ch,String wave_filename,int16_t freq) {
  wave_name[ch] = wave_filename;
  wave_order[ch] = WAVE_MODE_RUN_START;
  wave_next_name[ch] = "none";
}

void play_wav_repeat(int8_t ch,String wave_filename,int16_t freq) {
  wave_name[ch] = wave_filename;
  wave_order[ch] = WAVE_MODE_RUN_START;
  wave_next_name[ch] = "aaaa";
}

void stop_wav_all() {
  for(int i = 0;i < WAVE_CHORDS;i++){
    stop_wav(i);
  }
}

void stop_wav(int8_t ch) {
  wave_order[ch] = WAVE_MODE_STOP;
  wave_next_name[ch] = "none";
}


void serialEvent() {
  //digitalWrite(PIN_DEBUG_MANAGE,HIGH);
  while (Serial.available()) {
    // get the new byte:
    char inChar = (char)Serial.read();
    if (inChar == '\n') {
      stringComplete = true;
      //digitalWrite(PIN_DEBUG_MANAGE,LOW);
      return;
    }
    // add it to the inputString:
    inputString += inChar;
  }
  
}

void setup() {
  // put your setup code here, to run once:
  //ITimer1.attachInterruptInterval(TIMER1_INTERVAL_us, TimerHandler1);
  gpio_set_function(PIN_BUZZER, GPIO_FUNC_PWM);
  slice_num = pwm_gpio_to_slice_num(PIN_BUZZER);
  pwm_set_wrap(slice_num, 1023);
  pwm_set_chan_level(slice_num, PWM_CHAN_A, 0);
  //pwm_set_chan_level(slice_num, PWM_CHAN_B, 0);

  //debug pin 
  pinMode(PIN_DEBUG_PWM, OUTPUT);
  pinMode(PIN_DEBUG_MANAGE, OUTPUT);
  digitalWrite(PIN_DEBUG_PWM,LOW);
  digitalWrite(PIN_DEBUG_MANAGE,LOW);

  LittleFS.begin();

  //初期化
  for(int i = 0;i < WAVE_CHORDS;i++) {
    wave_position[i] = 0;
    wave_length[i] = 0;
    wave_volume[i] = 255;
    wave_last_val[i] = 127;
    //wave_last_val[i] = 0;
    wave_status[i] = WAVE_MODE_STOP;
    wave_order[i] = WAVE_MODE_NONE;
    wave_name[i] = "none";
    wave_next_name[i] = "none";
  }

  //当該バッファを０リセット
  for(int i = 0; i < BUF_LENGTH;i++){
    wave_buf_a[i] = 0;
    wave_buf_b[i] = 0;
  }
  Serial.begin(115200);
  inputString.reserve(200);
        
  pwm_set_enabled(slice_num, true);
  //Timer開始
  add_repeating_timer_us( -1 * TIMER1_INTERVAL_us, &timer_callback, NULL, &timer );
}

void loop() {
  // put your main code here, to run repeatedly:
  if (stringComplete) {
    Serial.println(inputString);
    if(inputString.equals("0")){play_wav(0,"185_CP_C2000.wav",16);}
    else if(inputString.equals("1")){play_wav(1,"185_door_open2.wav",16);}
    else if(inputString.equals("2")){play_wav(2,"185_shanai_chime.wav",16);}
    else if(inputString.equals("3")){play_wav(3,"185_shinagawa_B.wav",16);}
    else if(inputString.equals("4")){play_wav(4,"CrossingCountry.wav",16);}
    else if(inputString.equals("5")){play_wav(5,"Station_Announce01.wav",16);}
    else if(inputString.equals("6")){play_wav(6,"SineWave2.wav",16);}
    else if(inputString.equals("7")){play_wav(7,"SawtoothWave.wav",16);}
    else if(inputString.equals("r0")){play_wav_repeat(1,"185_CP_C2000.wav",16);}
    else if(inputString.equals("r1")){play_wav_repeat(1,"185_door_open2.wav",16);}
    else if(inputString.equals("r2")){play_wav_repeat(2,"185_shanai_chime.wav",16);}
    else if(inputString.equals("r3")){play_wav_repeat(3,"185_shinagawa_B.wav",16);}
    else if(inputString.equals("r4")){play_wav_repeat(4,"CrossingCountry.wav",16);}
    else if(inputString.equals("r5")){play_wav_repeat(5,"Station_Announce01.wav",16);}
    else if(inputString.equals("r6")){play_wav_repeat(6,"SineWave2.wav",16);}
    else if(inputString.equals("r7")){play_wav_repeat(7,"SawtoothWave.wav",16);}
    else if(inputString.equals("c")){stop_wav_all();}
    else if(inputString.equals("c0")){stop_wav(0);}
    else if(inputString.equals("c1")){stop_wav(1);}
    else if(inputString.equals("c2")){stop_wav(2);}
    else if(inputString.equals("c3")){stop_wav(3);}
    else if(inputString.equals("c4")){stop_wav(4);}
    else if(inputString.equals("c5")){stop_wav(5);}
    else if(inputString.equals("c6")){stop_wav(6);}
    else if(inputString.equals("c7")){stop_wav(7);}
    // clear the string:
    inputString = "";
    stringComplete = false;
  }
  manage_wave();

  
}
